"""
Constants and smart defaults for use with modelling-service
"""

"""
(min, max) range of hyperparams for building param grid in bayesian optimization
"""
bounds = dict(
    # Logistic Regression Classifier
    lrc = {
        'regParam': (0.1, 1),
        'elasticNetParam': (0.1, 1),
        'threshold': (0.4, 0.6)
    },

    # Decision Tree Classifier
    dtc = {
        'maxDepth': (1,10),
        'minInstancesPerNode': (1,10),
        'minInfoGain':(0.0,0.5),
        'maxBins':(32,64)
    },

    # Linear Support Vector Classifier
    # lsvc = {
    #     'regParam': (0.1,1.0)
    # },

    # Gradient Boosted Tree Classifier
    gbtc = {
        'maxDepth': (1,10),
        'minInstancesPerNode': (1,10),
        'minInfoGain': (0.0,0.5),
        'maxBins': (32,64),
        'stepSize': (0.1,1.0),
        'subsamplingRate': (0.5,1.0)
    },

    # Random Forest Classifier
    rfc = {
        'maxDepth': (1,10),
        'minInstancesPerNode': (1,10),
        'minInfoGain': (0.0,0.5),
        'maxBins': (32,64),
        'numTrees': (10,50),
        'subsamplingRate':(0.5,1.0)
    },

    #Naive Bayes Classifier
    nbc = {
        'smoothing': (0.1,1)
    }
)

"""
Smart default hyperparams. Generates param grid by cross product
"""
inits = dict(
    # Logistic Regression Classifier
    lrc = {
        'regParam': [0.1, 0.3],
        'elasticNetParam': [0.1, 0.3],
        'threshold': [0.4, 0.5]
    },

    #Decision Tree Classifier
    dtc = {
        'maxDepth': [1,3],
        'minInstancesPerNode': [1,3],
        'minInfoGain': [0.0,0.1],
        'maxBins': [32,40]
    },

    #Linear Support Vector Classifier
    # lsvc = {
    #     'regParam': [0.1,0.3]
    # },

    #Gradient Boosted Tree Classifier
    gbtc = {
        'maxDepth': [1,3],
        'minInstancesPerNode': [1,3],
        'minInfoGain': [0.0,0.1],
        'maxBins': [32,40],
        'stepSize': [0.1,0.5],
        'subsamplingRate': [0.5,1.0]
    },

    #Random Forest Classifier
    rfc = {
        'maxDepth': [1,3],
        'minInstancesPerNode': [1,3],
        'minInfoGain': [0.0,0.1],
        'maxBins': [32,40],
        'numTrees': [20,40],
        'subsamplingRate': [0.5,1.0]
    },

    #Naive Bayes Classifier
    nbc = {
        'smoothing': [0.5,1.0]
    }
)

"""
Dictionary containing the default hyperparams set by Spark
"""
defaults = dict(
    # Logistic Regression Classifier
    lrc = {
        'regParam': [0.0],
        'elasticNetParam': [0.0],
        'threshold': [0.5]
    },
    
    #Decision Tree Classifier
    dtc = {
        'maxDepth': [5],
        'minInstancesPerNode': [1],
        'minInfoGain': [0.0],
        'maxBins': [32]
    },

    #Linear Support Vector Classifier
    # lsvc = {
    #     'regParam': [0.0]
    # },

    #Gradient Boosted Tree Classifier
    gbtc = {
        'maxDepth': [5],
        'minInstancesPerNode': [1],
        'minInfoGain': [0.0],
        'maxBins': [32],
        'stepSize': [0.1],
        'subsamplingRate': [1.0]
    },

    #Random Forest Classifier
    rfc = {
        'maxDepth': [5],
        'minInstancesPerNode': [1],
        'minInfoGain': [0.0],
        'maxBins': [32],
        'numTrees': [20],
        'subsamplingRate': [1.0]
    },

    #Naive Bayes Classifier
    nbc = {
        'smoothing': [1.0]
    }
)


import numpy as np
import uuid
import itertools

from bayes_opt import BayesianOptimization

from modellingService.Utils import generateDirName, getEstimator, getEvaluator, getParamGrid, getValidator, getScorer, \
    reshapeParams, getParamInits, getDefaultParams, getParamBounds, generateReport

import Estimators as e

def train(spark, inputDF, algorithm, hyperparams={}, metric='areaUnderROC', numFolds=3, seed=None, numInitPoints=5, numIterations=2, acquisitionFunction='poi', gpParams={'alpha': 1e-5}, baseDir='output', workDir='output', outputDir=uuid.uuid4(), idx=1, binary=True, smart=False, optimize=False, persist=True, s3=False):
    """
    A single point of entry for both single algorithm and bayesian optimization
    :param spark: an existing spark session (unused for now)
    :param inputDF: dataframe containing features and labels
    :param algorithm: the algorithm code
    :param hyperparams: a dictionary of custom param overrides
    :param metric: scoring metric to be used
    :param numFolds: kFold value for cross validation
    :param seed: a seed value to set for deterministic debugging and testing
    :param numInitPoints: number of initial points to explore for bayesian optimization
    :param numIterations: number of iterations to exploit for bayesian optimization
    :param acquisitionFunction: acquisition function to use for bayesian optimization
    :param gpParams: params for tuning bayesian optimization
    :param baseDir: the directory where np will save report.csv 
    :param outputDir: the directory name for the model training job
    :param idx: the index of the model being trained in the current job
    :param binary: is the data multi-class of binary
    :param smart: if you want to explore the set of initial points set by the data scientists
    :param optimize: if you want to run bayesian optimization
    :param persist: if you want to persist model and report output
    :return: A JSON report of the trained model
    """
    _counter = itertools.count()
    _dirName = '{}/{}'.format(outputDir, generateDirName(algorithm, optimize, idx))
    estimator = getEstimator(algorithm)(hyperparams)
    evaluator = getEvaluator(metric, binary)

    def getWrapperFunction():
        def optimizer(**params):
            paramGrid = getParamGrid(estimator, params)
            validator = getValidator(estimator, evaluator, paramGrid, numFolds=numFolds, seed=seed)
            model = validator.fit(inputDF)

            if (persist):
                model.bestModel.save('{}/{}/{}'.format(baseDir, _dirName, _counter.next() + 1))

            if (optimize):
                score = model.avgMetrics
                scorer = getScorer(algorithm)
                return scorer(score)
            else:
                return model
        return optimizer

    # ensure the custom params are cleaned first
    customParams = reshapeParams(hyperparams, e.reshapers[algorithm])
    
    # fills in missing params with either smart or default params
    if (smart | optimize):
        smartParams = getParamInits(algorithm, customParams)
    else:
        smartParams = getDefaultParams(algorithm, customParams)

    f = getWrapperFunction()

    if (optimize):
        paramBounds = getParamBounds(algorithm, customParams)
        bo = BayesianOptimization(f, paramBounds, verbose=0)
        bo.explore(smartParams)
        bo.maximize(init_points=numInitPoints, n_iter=numIterations, acq=acquisitionFunction, **gpParams)

        report = generateReport('{}/{}'.format(baseDir, _dirName), workDir, algorithm, smartParams, bo.space.keys, bo.space.X, bo.space.Y, metric=metric, persist=persist, s3=s3)

        return report
    else:
        model = f(**smartParams)
        paramMap = model.getEstimatorParamMaps()[np.argmax(model.avgMetrics)]
        keys = list(map(lambda p: p.name, paramMap))
        X = [list(map(lambda p: paramMap[p], paramMap))]
        Y = [max(model.avgMetrics)]

        report = generateReport('{}/{}'.format(baseDir, _dirName), workDir, algorithm, smartParams, keys, X, Y, metric=metric, persist=persist, s3=s3)

        return report


from pyspark.ml.classification import LogisticRegression, DecisionTreeClassifier, GBTClassifier, RandomForestClassifier, NaiveBayes
import Utils

"""
Each estimator wrapper does the following:
- take a free form dict of hyperparams as input
- filters out invalid params using reshaper
- passes clean params to Spark estimator
"""

def logisticRegressionClassifier(hyperparams = {}, reshape=True):
    if (reshape):
        params = Utils.reshapeParams(hyperparams, reshapers['lrc'])
    else:
        params = hyperparams
    return LogisticRegression(**params)

def decisionTreeClassifier(hyperparams = {}, reshape=True):
    if (reshape):
        params = Utils.reshapeParams(hyperparams, reshapers['dtc'])
    else:
        params = hyperparams
    return DecisionTreeClassifier(**params)

# def linearSupportVectorClassifier(hyperparams = {}, reshape=True):
#     if (reshape):
#         params = Utils.reshapeParams(hyperparams, reshapers['lsvc'])
#     else:
#         params = hyperparams
#     return LinearSVC(**params)

def gradientBoosterTreeClassifier(hyperparams = {}, reshape=True):
    if (reshape):
        params = Utils.reshapeParams(hyperparams, reshapers['gbtc'])
    else:
        params = hyperparams
    return GBTClassifier(**params)

def randomForestClassifier(hyperparams = {}, reshape=True):
    if (reshape):
        params = Utils.reshapeParams(hyperparams, reshapers['rfc'])
    else:
        params = hyperparams
    return RandomForestClassifier(**params)

def naiveBayesClassifier(hyperparams = {}, reshape=True):
    if (reshape):
        params = Utils.reshapeParams(hyperparams, reshapers['nbc'])
    else:
        params = hyperparams
    return NaiveBayes(**params)

estimators = dict(
    lrc=logisticRegressionClassifier,
    dtc=decisionTreeClassifier,
    # lsvc=linearSupportVectorClassifier,
    gbtc=gradientBoosterTreeClassifier,
    rfc=randomForestClassifier,
    nbc=naiveBayesClassifier
)

"""
A reshaper is a map of cast functions to avoid precision errors in hyperparams
"""
reshapers = dict(
    lrc = {
        'regParam': float,
        'elasticNetParam': float,
        'threshold': float
    },
    dtc = {
        'maxDepth': int,
        'minInstancesPerNode': int,
        'maxBins': int,
        'minInfoGain': float
    },
    # lsvc = {
    #     'regParam': float
    # },
    gbtc = {
        'maxDepth': int,
        'minInstancesPerNode': int,
        'maxBins': int,
        'minInfoGain': float,
        'stepSize': float,
        'subsamplingRate': float
    },
    rfc = {
        'maxDepth': int,
        'minInstancesPerNode': int,
        'maxBins': int,
        'minInfoGain': float,
        'numTrees': int,
        'subsamplingRate': float
    },
    nbc = {
        'smoothing': float
    }
)

"""
A scorer determines how to pick the best score for each algorithm
"""
scorers = dict(
    lrc = lambda score: max(score),
    dtc = lambda score: max(score),
    # lsvc = lambda score: max(score),
    gbtc = lambda score: max(score),
    rfc = lambda score: max(score),
    nbc = lambda score: max(score)
)


from setuptools import setup

setup(
    name='modelling-service-test',
    version='0.1.4',
    packages=['modellingService'],
    install_requires=['bayesian-optimization'],
    author='CAPS-AN-EN',
    author_email='caps-an-en@mediaiqdigital.com',
    description='A pyspark package for modelling-service algorithms'
)



import numpy as np
import itertools
import json
import uuid
import time
import datetime

from pyspark.ml.evaluation import BinaryClassificationEvaluator, MulticlassClassificationEvaluator
from pyspark.ml.tuning import ParamGridBuilder, CrossValidator

import Estimators as e
import Constants as c

def getInputDF(spark, inputPath, inputFormat):
    """
    Returns DataFrame from input configurations
    :param spark:
    :param inputPath:
    :param inputFormat:
    :return: DataFrame
    """
    df = spark.read.format(inputFormat).load(inputPath)
    return df

def reshapeParams(hyperparams, reshaper):
    """
    A utility function that takes a dictionary of transformations for valid hyperparameters
    and filters out invalid params

    Testing the function:
    print reshapeParams({'regParam': 'invalid', 'elasticNetParam': 0.1, 'maxDepth': 0.1}, {'regParam': float, 'elasticNetParam': float, 'threshold': float})
    # Expect: {'elasticNetParam': 0.1}
    :param hyperparams: the dirty dictionary
    :param reshaper: the transformations
    :return: the clean dictionary
    """
    d = {}
    for k, v in hyperparams.items():
        if (reshaper.has_key(k)):
            f = reshaper[k]
            if (isinstance(v, str)):
                if (v.replace('.', '', 1).isdigit()):
                    d[k] = f(v)
                else:
                    continue
            elif (isinstance(v, float)):
                d[k] = f(v)
            elif (isinstance(v, int)):
                d[k] = f(v)
            else:
                continue
        else:
            continue
    return d

def getEstimator(algorithm):
    """
    Returns a wrapper to a valid Estimator and throws an exception if unimplemented
    :param algorithm: the algorithm code
    :return:
    """
    if (e.estimators.has_key(algorithm)):
        return e.estimators[algorithm]
    else:
        raise Exception('{} has not been implemented yet, try one of {}'.format(algorithm, e.estimators.keys()))

def getEvaluator(metric='areaUnderROC', binary=True):
    """
    Returns an evaluator for classification algorithms based on Binary or Multiclass data
    :param metric: For binary - areaUnderROC|areaUnderPR, For multiclass - f1|weightedPrecision|weightedRecall|accuracy
    :param binary: Boolean
    :return: Evaluator object
    """
    if (binary):
        return BinaryClassificationEvaluator(metricName=metric)
    else:
        return MulticlassClassificationEvaluator(metricName=metric)

def getParamBounds(algorithm, customParams):
    """
    Overrides custom params over init points
    :param algorithm:
    :param customParams: Expects params that have been reshaped
    :return: Bounds of hyperparams
    """
    if (c.bounds.has_key(algorithm)):
        points = c.bounds[algorithm].copy()
        for k, v in points.items():
            if (customParams.has_key(k)):
                points.pop(k)
        return points
    else:
        raise Exception('{} does not have paramBound defined yet'.format(algorithm))

def getParamInits(algorithm, customParams):
    """
    Overrides custom params over init points
    :param algorithm:
    :param customParams: Expects params that have been reshaped
    :return: Set of hyperparams
    """
    if (c.inits.has_key(algorithm)):
        points = c.inits[algorithm].copy()
        for k, v in points.items():
            if (customParams.has_key(k)):
                points[k] = [customParams[k]]
        return points
    else:
        raise Exception('{} does not have paramInit defined yet'.format(algorithm))

def getDefaultParams(algorithm, customParams):
    """
    Fetches default params
    :param algorithm: 
    :param customParams: 
    :return: 
    """
    if (c.defaults.has_key(algorithm)):
        defaults = c.defaults[algorithm].copy()
        for k, v in defaults.items():
            if (customParams.has_key(k)):
                defaults[k] = [customParams[k]]
        return defaults
    else:
        raise Exception('{} does not have defaults defined yet'.format(algorithm))

def getScorer(algorithm):
    """
    Fetches the scorer function for the algorithm
    :param algorithm: 
    :return: 
    """
    if (e.scorers.has_key(algorithm)):
        return e.scorers[algorithm]
    else:
        raise Exception('{} does not have scorer defined yet'.format(algorithm))

def getParamGrid(estimator, params):
    """
    Generates a grid of hyperparams based on param bounds
    :param estimator: estimator for extracting the hyperparams
    :param params: the bounds for each hyperparam
    :return: ParamGrid object
    """
    paramGrid = ParamGridBuilder()
    for p in params:
        if (type(params[p]) == list):
            paramGrid = paramGrid.addGrid(getattr(estimator, p), params[p])
        else:
            paramGrid = paramGrid.addGrid(getattr(estimator, p), [params[p]])
    return paramGrid.build()

def getValidator(estimator, evaluator, paramGrid, numFolds=3, seed=42):
    """
    Creates a CrossValidator object
    :param estimator:
    :param evaluator:
    :param paramGrid:
    :param numFolds:
    :param seed:
    :return: Returns a CrossValidator object
    """
    validator = CrossValidator(estimator=estimator, evaluator=evaluator, estimatorParamMaps=paramGrid, numFolds=numFolds, seed=seed)
    return validator

def getSignificantDigits(values):
    """
    Utility to fix precision of float values
    :param values: 
    :return: 
    """
    return list(map(lambda v: float('{0:.4f}'.format(float(v))), values))

def generateModelReport(point, keys, modelDir):
    """
    Generates a dict for each model
    :param point: an ndarray containing hyperparam values and other metadata
    :param keys: headers to map the values to
    :param modelDir: output metadata
    :return: 
    """
    points = map(lambda v: v.item(), np.nditer(point))
    hyperparams = {}
    metrics = {}
    output = {'model_id': points[0], 'model_dir': modelDir, 'model_target': points[1]}
    for (k, v) in zip(keys, points)[-1:]:
        metrics[k] = v
    for (k, v) in zip(keys, points)[2:-1]:
        hyperparams[k] = v
    report = {'_id': points[0], 'output': output, 'hyperparams': hyperparams, 'metrics': metrics}
    return report

def generateReport(outputDir, workDir, algorithm, customParams, headers, params, values, metric='areaUnderROC', persist=True, s3=False):
    """
    Generates a JSON for each model run
    :param outputDir: directory to save the report
    :param algorithm: string code of the algorithm
    :param customParams: customParams taken here to fill in when optimized only on certain params
    :param headers: the params used for training model
    :param params: a matrix of param values across runs
    :param values: a vector of metric values across runs
    :param metric: the scoring metric used
    :param persist: if a csv needs to be written out
    :return: 
    """
    indexes = range(1, len(params) + 1)
    models = list(map(lambda idx: str('{}/{}'.format(outputDir, idx)), indexes))

    smartParams = {}
    for k, v in customParams.items():
        if k not in headers:
            smartParams[k] = v[0]

    points = np.hstack((np.expand_dims(indexes, axis=1), np.expand_dims(models, axis=1), list(map(lambda d: getSignificantDigits(d) + smartParams.values(), params)), np.expand_dims(getSignificantDigits(values), axis=1)))
    header = ','.join(['idx', 'model'] + headers + smartParams.keys() + [metric])

    modelDir = outputDir.split('/')[-1]
    jobName = outputDir.split('/')[-2]

    if (persist):
        if (s3):
            np.savetxt('{}/{}-{}-report.csv'.format(workDir, jobName, modelDir), points, header=header, delimiter=',', fmt='%s', comments='')
        else:
            np.savetxt('{}/report.csv'.format(outputDir), points, header=header, delimiter=',', fmt='%s', comments='')

    models = list(map(lambda p: generateModelReport(p, header.split(','), modelDir), points))
    report = json.dumps({'algorithm': algorithm, 'models': models}, sort_keys=True)

    return report

def generateDirName(algorithm, optimize=False, idx=1):
    """
    Utility function to generate a directory name as <algorithm>-<?optimize>-<idx>-<timestamp>
    :param algorithm: 
    :param optimize: 
    :param idx: 
    :return: 
    """
    tags = [algorithm]
    if (optimize):
        tags.append('optimized')
    now = time.time()
    tags.append(str(idx))
    return '-'.join(tags)


